@@ -5,24 +5,44 @@ import base64
 from datetime import datetime
 from hashlib import sha256
 import csv
+from io import BytesIO
+from zipfile import ZipFile
+from urllib.request import urlopen
+import pandas as pd
 
 requests.packages.urllib3.util.connection.HAS_IPV6 = False
 
-class SocketEventBreeze(socketio.ClientNamespace):
 
+urlmapper = dict()
+urlmapper['isec_script']  = "https://directlink.icicidirect.com/NewSecurityMaster/SecurityMaster.zip"
+urlmapper['order_feeds'] = "https://livefeeds.icicidirect.com"
+urlmapper['rate_refresh_feeds'] = "https://livestream.icicidirect.com"
+urlmapper['customer_detail_url'] = "https://api.icicidirect.com/breezeapi/api/v1/customerdetails"
+urlmapper['hostname'] = "https://api.icicidirect.com/breezeapi/api/v1/"
+urlmapper['stockscriptlist'] = "https://traderweb.icicidirect.com/Content/File/txtFile/ScripFile/StockScriptNew.csv"
+
+resp = urlopen(urlmapper['isec_script'])
+zipfile = ZipFile(BytesIO(resp.read()))
+
+
+class SocketEventBreeze(socketio.ClientNamespace):
+    
     def __init__(self, namespace, breeze_instance):
         super().__init__(namespace)
         self.breeze = breeze_instance
-        self.hostname = 'https://livestream.icicidirect.com'
         self.sio = socketio.Client()
 
-    def connect(self):
+    def connect(self,hostname):
         auth = {"user": self.breeze.user_id, "token": self.breeze.session_key}
-        self.sio.connect(self.hostname, headers={"User-Agent": "python-socketio[client]/socket"}, auth=auth, transports="websocket", wait_timeout=3)
-
+        socket = self.sio.connect(hostname, headers={"User-Agent": "python-socketio[client]/socket"},auth=auth,transports="websocket", wait_timeout=3)
+        print("Connected for hostname: "+str(hostname)+" socket object : "+str(self.sio))
+        
     def on_disconnect(self):
         self.sio.emit("disconnect", "transport close")
-
+        
+    def notify(self):
+        self.sio.on('order', self.on_message)
+        
     def on_message(self, data):
         data = self.breeze.parse_data(data)
         if 'symbol' in data and data['symbol'] != None and len(data['symbol'])>0:
@@ -35,31 +55,100 @@ class SocketEventBreeze(socketio.ClientNamespace):
 
     def unwatch(self, data):
         self.sio.emit("leave", data)
+        
 
 
 class BreezeConnect():
 
-    def __init__(self, api_key):  # needed for hashing json data
+    def __init__(self, api_key):  
         self.user_id = None
         self.api_key = api_key
         self.session_key = None
         self.secret_key = None
-        self.sio_handler = None
+        self.sio_rate_refresh_handler = None
+        self.sio_order_refresh_handler = None
         self.api_handler = None
         self.on_ticks = None
         self.stock_script_dict_list = []
         self.token_script_dict_list = []
+        self.tux_to_user_value = {
+            "orderFlow": {
+                "B": "Buy",
+                "S": "Sell",
+                "N": "NA"
+            },
+            "limitMarketFlag": {
+                "L": "Limit",
+                "M": "Market",
+                "S": "StopLoss"
+            },
+            "orderType": {
+                "T": "Day",
+                "I": "IoC",
+                "V": "VTC"
+            },
+            "productType": {
+                "F": "Futures",
+                "O": "Options",
+                "P": "FuturePlus",
+                "U": "FuturePlus_sltp",
+                "I": "OptionPlus",
+                "C": "Cash",
+                "Y": "eATM",
+                "B": "BTST",
+                "M": "Margin",
+                "T": "MarginPlus"
+            },
+            "orderStatus": {
+                "A": "All",
+                "R": "Requested",
+                "Q": "Queued",
+                "O": "Ordered",
+                "P": "Partially Executed",
+                "E": "Executed",
+                "J": "Rejected",
+                "X": "Expired",
+                "B": "Partially Executed And Expired",
+                "D": "Partially Executed And Cancelled",
+                "F": "Freezed",
+                "C": "Cancelled"
+            },
+            "optionType": {
+                "C": "Call",
+                "P": "Put",
+                "*": "Others"
+            },
+        }
 
+    def _ws_connect(self,handler,order_flag): 
+        if order_flag:
+            if not self.sio_order_refresh_handler:
+                self.sio_order_refresh_handler = SocketEventBreeze("/", self)
+            self.sio_order_refresh_handler.connect(urlmapper['order_feeds'])
+        else:
+            if not self.sio_rate_refresh_handler: 
+                self.sio_rate_refresh_handler = SocketEventBreeze("/", self)
+            self.sio_rate_refresh_handler.connect(urlmapper['rate_refresh_feeds'])
+    
+    def ws_disconnect(self,isOrder = False):
+        if(isOrder == False):    
+            if not self.sio_rate_refresh_handler:
+                return {"socket server is not connected to rate refresh."}
+            else:
+                self.sio_rate_refresh_handler.on_disconnect()
+                self.sio_rate_refresh_handler = None
+                return {"socket server for rate refresh  has been disconnected."}
+        else:
+            if not self.sio_order_refresh_handler:
+                return {"socket server is not connected to order refresh."}
+            else:    
+                self.sio_order_refresh_handler.on_disconnect()
+                self.sio_order_refresh_handler = None
+                return {"socket server for order streaming has been disconnected."}
+    
     def ws_connect(self):
-        if not self.sio_handler:
-            self.sio_handler = SocketEventBreeze("/", self)
-            self.sio_handler.connect()
-
-    def ws_disconnect(self):
-        if not self.sio_handler:
-            self.sio_handler = SocketEventBreeze("/", self)
-        self.sio_handler.on_disconnect()
-
+        self._ws_connect(self.sio_rate_refresh_handler,False)
+        
     def get_data_from_stock_token_value(self, input_stock_token):
         try:
             output_data = {}
@@ -176,32 +265,45 @@ class BreezeConnect():
                     market_depth_token_value = exchange_code_name + "2!" + token_value
                 return exchange_quotes_token_value, market_depth_token_value
 
-    def subscribe_feeds(self, stock_token="", exchange_code="", stock_code="", product_type="", expiry_date="", strike_price="", right="", get_exchange_quotes=True, get_market_depth=True):
-        if self.sio_handler:
+    def subscribe_feeds(self, stock_token="", exchange_code="", stock_code="", product_type="", expiry_date="", strike_price="", right="", get_exchange_quotes=True, get_market_depth=True, get_order_notification=False):
+        if self.sio_rate_refresh_handler:
             return_object = {}
+            if get_order_notification == True:
+                self._ws_connect(self.sio_order_refresh_handler,True)
+                self.sio_order_refresh_handler.notify()
+                return_object = {"message": "Order Notification subscribed successfully"}
             if stock_token != "":
-                self.sio_handler.watch(stock_token)
+                self.sio_rate_refresh_handler.watch(stock_token)
                 return_object = {"message": "Stock " + stock_token + " subscribed successfully"}
+            elif get_order_notification == True and exchange_code == "":
+                return return_object
             else:
                 exchange_quotes_token, market_depth_token = self.get_stock_token_value(exchange_code=exchange_code, stock_code=stock_code, product_type=product_type, expiry_date=expiry_date, strike_price=strike_price, right=right, get_exchange_quotes=get_exchange_quotes, get_market_depth=get_market_depth)
                 if exchange_quotes_token != False:
-                    self.sio_handler.watch(exchange_quotes_token)
+                    self.sio_rate_refresh_handler.watch(exchange_quotes_token)
                 if market_depth_token != False:
-                    self.sio_handler.watch(market_depth_token)
+                    self.sio_rate_refresh_handler.watch(market_depth_token)
                 return_object = {"message": "Stock " + stock_code + " subscribed successfully"}
             return return_object
-
-    def unsubscribe_feeds(self, stock_token="", exchange_code="", stock_code="", product_type="", expiry_date="", strike_price="", right="", get_exchange_quotes=True, get_market_depth=True):
-        if self.sio_handler:
+        
+    def unsubscribe_feeds(self, stock_token="", exchange_code="", stock_code="", product_type="", expiry_date="", strike_price="", right="", get_exchange_quotes=True, get_market_depth=True,get_order_notification=False):
+        if(get_order_notification == True):
+            if self.sio_order_refresh_handler:
+                self.sio_order_refresh_handler.on_disconnect()
+                self.sio_order_refresh_handler = None
+                return {"socket server for order streaming has been disconnected."}
+            else:
+                return {"order streaming server is not connected."}
+        if self.sio_rate_refresh_handler:
             if stock_token != "":
-                self.sio_handler.unwatch(stock_token)
+                self.sio_rate_refresh_handler.unwatch(stock_token)
                 return {"message": "Stock " + stock_token + " unsubscribed successfully"}
             else:
                 exchange_quotes_token, market_depth_token = self.get_stock_token_value(exchange_code=exchange_code, stock_code=stock_code, product_type=product_type, expiry_date=expiry_date, strike_price=strike_price, right=right, get_exchange_quotes=get_exchange_quotes, get_market_depth=get_market_depth)
                 if exchange_quotes_token != False:
-                    self.sio_handler.unwatch(exchange_quotes_token)
+                    self.sio_rate_refresh_handler.unwatch(exchange_quotes_token)
                 if market_depth_token != False:
-                    self.sio_handler.unwatch(market_depth_token)
+                    self.sio_rate_refresh_handler.unwatch(market_depth_token)
                 return {"message": "Stock " + stock_code + " unsubscribed successfully"}
 
     def parse_market_depth(self, data, exchange):
@@ -423,7 +525,7 @@ class BreezeConnect():
                 "AppKey": self.api_key
             }
             body = json.dumps(body, separators=(',', ':'))
-            url = "https://api.icicidirect.com/breezeapi/api/v1/customerdetails"
+            url = urlmapper['customer_detail_url']
             response = requests.get(url=url, data=body, headers=headers)
             if response.json()['Success'] != None:
                 base64_session_token = response.json()['Success']['session_token']
@@ -440,7 +542,7 @@ class BreezeConnect():
             self.stock_script_dict_list = [{}, {}, {}, {}, {}]
             self.token_script_dict_list = [{}, {}, {}, {}, {}]
             with requests.Session() as s:
-                download = s.get("https://traderweb.icicidirect.com/Content/File/txtFile/ScripFile/StockScriptNew.csv")
+                download = s.get(urlmapper['stockscriptlist'])
                 decoded_content = download.content.decode('utf-8')
                 cr = csv.reader(decoded_content.splitlines(), delimiter=',')
                 my_list = list(cr)
@@ -545,13 +647,17 @@ class BreezeConnect():
     def get_trade_detail(self, exchange_code="", order_id=""):
         if self.api_handler:
             return self.api_handler.get_trade_detail(exchange_code, order_id)
+    
+    def get_names(self,exchange_code="",stock_code=""):
+        if self.api_handler:
+            return self.api_handler.get_names(exchange_code, stock_code)
 
 
 class ApificationBreeze():
 
     def __init__(self, breeze_instance):
         self.breeze = breeze_instance
-        self.hostname = 'https://api.icicidirect.com/breezeapi/api/v1/'
+        self.hostname = urlmapper['hostname']
         self.base64_session_token = base64.b64encode(
             (self.breeze.user_id + ":" + self.breeze.session_key).encode('ascii')).decode('ascii')
 
@@ -569,7 +675,7 @@ class ApificationBreeze():
             }
             return headers
         except Exception as e:
-            print("generate_headers() Error - ", e)
+            raise Exception("generate_headers() Error - " , e)
 
     def make_request(self, method, endpoint, body, headers):
         try:
@@ -589,7 +695,7 @@ class ApificationBreeze():
             else:
                 print("Invalid Request Method - Must be GET, POST, PUT or DELETE")
         except Exception as e:
-            print("Error while trying to make request "+method+" "+url+" - ", e)
+            raise Exception("Error while trying to make request "+method+" "+url+" - ", e)
 
     def get_customer_details(self, api_session=""):
         try:
@@ -611,7 +717,7 @@ class ApificationBreeze():
                 del response['Success']['session_token']
             return response
         except Exception as e:
-            print("get_customer_details() Error - ", e)
+            raise Exception("get_customer_details() Error - ", e)
 
     def get_demat_holdings(self):
         try:
@@ -622,7 +728,7 @@ class ApificationBreeze():
             response = response.json()
             return response
         except Exception as e:
-            print("get_demat_holdings() Error- ", e)
+            raise Exception("get_demat_holdings() Error- ", e)
 
     def get_funds(self):
         try:
@@ -633,7 +739,7 @@ class ApificationBreeze():
             response = response.json()
             return response
         except Exception as e:
-            print("get_funds() Error - ", e)
+            raise Exception("get_funds() Error - ", e)
 
     def set_funds(self, transaction_type="", amount="", segment=""):
         try:
@@ -663,7 +769,7 @@ class ApificationBreeze():
             response = response.json()
             return response
         except Exception as e:
-            print("set_funds() Error - ", e)
+            raise Exception("set_funds() Error - ", e)
 
     def get_historical_data(self, interval="", from_date="", to_date="", stock_code="", exchange_code="", product_type="", expiry_date="", right="", strike_price=""):
         try:
@@ -727,7 +833,7 @@ class ApificationBreeze():
             response = response.json()
             return response
         except Exception as e:
-            print("get_historical_data() Error - ", e)
+            raise Exception("get_historical_data() Error - ", e)
 
     def add_margin(self, product_type="", stock_code="", exchange_code="", settlement_id="", add_amount="", margin_amount="", open_quantity="", cover_quantity="", category_index_per_stock="", expiry_date="", right="", contract_tag="", strike_price="", segment_code=""):
         try:
@@ -775,7 +881,7 @@ class ApificationBreeze():
             response = response.json()
             return response
         except Exception as e:
-            print("add_margin() Error - ", e)
+            raise Exception("add_margin() Error - ", e)
 
     def get_margin(self, exchange_code=""):
         try:
@@ -791,7 +897,7 @@ class ApificationBreeze():
             response = response.json()
             return response
         except Exception as e:
-            print("get_margin() Error - ", e)
+            raise Exception("get_margin() Error - ", e)
 
     def place_order(self, stock_code="", exchange_code="", product="", action="", order_type="", stoploss="", quantity="", price="", validity="", validity_date="", disclosed_quantity="", expiry_date="", right="", strike_price="", user_remark=""):
         try:
@@ -862,7 +968,7 @@ class ApificationBreeze():
             response = response.json()
             return response
         except Exception as e:
-            print("place_order() Error - ", e)
+            raise Exception("place_order() Error - ", e)
 
     def get_order_detail(self, exchange_code, order_id):
         try:
@@ -883,7 +989,7 @@ class ApificationBreeze():
             response = response.json()
             return response
         except Exception as e:
-            print("get_order_detail() Error - ", e)
+            raise Exception("get_order_detail() Error - ", e)
 
     def get_order_list(self, exchange_code, from_date, to_date):
         try:
@@ -908,7 +1014,7 @@ class ApificationBreeze():
             response = response.json()
             return response
         except Exception as e:
-            print("get_order_list() Error - ", e)
+            raise Exception("get_order_list() Error - ", e)
 
     def cancel_order(self, exchange_code, order_id):
         try:
@@ -929,7 +1035,7 @@ class ApificationBreeze():
             response = response.json()
             return response
         except Exception as e:
-            print("cancel_order() Error - ", e)
+            raise Exception("cancel_order() Error - ", e)
 
     def modify_order(self, order_id, exchange_code, order_type, stoploss, quantity, price, validity, disclosed_quantity, validity_date):
         try:
@@ -970,7 +1076,7 @@ class ApificationBreeze():
             response = response.json()
             return response
         except Exception as e:
-            print("modify_order() Error - ", e)
+            raise Exception("modify_order() Error - ", e)
 
     def get_portfolio_holdings(self, exchange_code, from_date, to_date, stock_code, portfolio_type):
         try:
@@ -995,7 +1101,7 @@ class ApificationBreeze():
             response = response.json()
             return response
         except Exception as e:
-            print("get_portfolio_holdings() Error - ", e)
+            raise Exception("get_portfolio_holdings() Error - ", e)
 
     def get_portfolio_positions(self):
         try:
@@ -1007,7 +1113,7 @@ class ApificationBreeze():
             response = response.json()
             return response
         except Exception as e:
-            print("get_portfolio_positions() Error - ", e)
+            raise Exception("get_portfolio_positions() Error - ", e)
 
     def get_quotes(self, stock_code, exchange_code, expiry_date, product_type, right, strike_price):
         try:
@@ -1042,7 +1148,7 @@ class ApificationBreeze():
             response = response.json()
             return response
         except Exception as e:
-            print("get_quotes() Error - ", e)
+            raise Exception("get_quotes() Error - ", e)
 
     def get_option_chain_quotes(self,stock_code, exchange_code, expiry_date, product_type, right, strike_price):
         try:
@@ -1052,13 +1158,13 @@ class ApificationBreeze():
             elif product_type=="" or product_type== None:
                 response = {"Success": "", "Status": 500, "Error": "Product-Type cannot be empty for Exchange-Code value as 'nfo'."}
                 return response
-            elif product_type!="future" and product_type!="options":
+            elif product_type.lower()!="futures" and product_type.lower()!="options":
                 response = {"Success": "", "Status": 500, "Error": "Product-type should be either 'futures' or 'options' for Exchange-Code value as 'nfo'." }
                 return response
             elif stock_code=="" or stock_code==None:
                 response = {"Success": "", "Status": 500, "Error": "stock code cannot be empty"}
                 return response
-            elif product_type == 'options':
+            elif product_type.lower() == 'options':
                 if((expiry_date=="" or expiry_date==None)  and (strike_price=="" or strike_price==None) and (right=="" or right==None)):
                     response = {"Success": "", "Status": 500, "Error": "Atleast two inputs are required out of Expiry-Date, Right & Strike-Price. All three cannot be empty'."}
                     return response
@@ -1071,7 +1177,7 @@ class ApificationBreeze():
                 elif((expiry_date=="" or expiry_date==None) and (strike_price=="" or strike_price==None) and (right!=None or right!="")):
                     response = {"Success": "", "Status": 500, "Error": "Either Expiry-Date or Strike-Price cannot be empty."}
                     return response
-                elif((right!="" and right!=None) and (right!="call" and right!="put" and right!="others")):
+                elif((right!="" and right!=None) and (right.lower()!="call" and right.lower()!="put" and right.lower()!="others")):
                     response = {"Success": "", "Status": 500, "Error": "Right should be either 'call', 'put', or 'others'."}
                     return response
             body = {
@@ -1091,8 +1197,8 @@ class ApificationBreeze():
             response = self.make_request("GET", "optionchain", body, headers)
             response = response.json()
             return response
-        except:
-            print("get_option_chain_quotes() Error -",e)
+        except Exception as e:
+            raise Exception("get_option_chain_quotes() Error -",e)
 
     def square_off(self, source_flag, stock_code, exchange_code, quantity, price, action, order_type, validity, stoploss, disclosed_quantity, protection_percentage, settlement_id, margin_amount, open_quantity, cover_quantity, product, expiry_date, right, strike_price, validity_date, trade_password, alias_name):
         try:
@@ -1126,7 +1232,7 @@ class ApificationBreeze():
             response = response.json()
             return response
         except Exception as e:
-            print("square_off() Error - ", e)
+            raise Exception("square_off() Error - ", e)
 
     def get_trade_list(self, from_date, to_date, exchange_code, product_type, action, stock_code):
         try:
@@ -1158,7 +1264,7 @@ class ApificationBreeze():
             response = response.json()
             return response
         except Exception as e:
-            print("get_trade_list() Error - ", e)
+            raise Exception("get_trade_list() Error - ", e)
 
     def get_trade_detail(self, exchange_code, order_id):
         try:
@@ -1178,4 +1284,42 @@ class ApificationBreeze():
             response = response.json()
             return response
         except Exception as e:
-            print("get_trade_detail() Error - ", e)
+            raise Exception("get_trade_detail() Error - ", e)
+    
+    def get_names(self, exchange_code, stock_code):
+        try:
+            lexchange_code = exchange_code.lower()
+            stock_code = stock_code.upper()
+            mapper_exchangecode_to_file = dict()
+            mapper_exchangecode_to_file['nse'] = 'NSEScripMaster.txt'
+            mapper_exchangecode_to_file['bse'] = 'BSEScripMaster.txt'
+            mapper_exchangecode_to_file['cdnse'] = 'CDNSEScripMaster.txt'
+            mapper_exchangecode_to_file['fonse'] = 'FONSEScripMaster.txt'
+            required_file = zipfile.open(mapper_exchangecode_to_file.get(lexchange_code))
+    
+            dataframe = pd.read_csv(required_file, sep=',', engine='python')
+             
+            df2 = dataframe[(dataframe[' "ExchangeCode"'] == stock_code) | (dataframe[' "ShortName"'] == stock_code)]
+            
+            if(len(df2)==0):
+                return({"Status ": "Result Not Found"})
+            requiredresult = df2[[' "ShortName"',' "ExchangeCode"','Token',' "CompanyName"']]
+    
+            isec_stock = requiredresult[' "ShortName"'].to_string().split()[1]
+            token = " ".join(requiredresult['Token'].to_string().split()[1:])
+            exchange = requiredresult[' "ExchangeCode"'].to_string()
+            compname = " ".join(requiredresult[' "CompanyName"'].to_string().split()[1:])
+            if(" " in exchange):
+                exchange = " ".join(exchange.split()[1:])
+            
+            result = {}
+            result['exchange_code'] = exchange_code
+            result['exchange_stock_code'] = exchange
+            result['isec_stock_code'] = isec_stock
+            result['isec_token'] = token
+            result['isec_token_level1'] = str('4.1!') + str(token)
+            result['isec_token_level2'] = str('4.2!') + str(token)
+            result['company name'] = compname                
+            return(result)
+        except Exception as e:
+            raise Exception({"Status" : "get_names() : data not found"})
